Block-Structured AMR Software Framework
Loading...
Searching...
No Matches
AMReX_EB2_Level.H
Go to the documentation of this file.
1#ifndef AMREX_EB2_LEVEL_H_
2#define AMREX_EB2_LEVEL_H_
3#include <AMReX_Config.H>
4
5#include <AMReX_ParmParse.H>
6#include <AMReX_Geometry.H>
7#include <AMReX_MultiFab.H>
8#include <AMReX_iMultiFab.H>
9#include <AMReX_LayoutData.H>
10#include <AMReX_VisMF.H>
11#include <AMReX_Array.H>
12#include <AMReX_EBCellFlag.H>
13#include <AMReX_MultiCutFab.H>
14#if (AMREX_SPACEDIM == 3)
15# include <AMReX_MarchingCubes.H>
16#endif
17#include <AMReX_EB2_MultiGFab.H>
18#include <AMReX_EB2_C.H>
20
21#ifdef AMREX_USE_OMP
22#include <omp.h>
23#endif
24
25#include <cmath>
26#include <limits>
27#include <map>
28#include <type_traits>
29#include <unordered_map>
30
31namespace amrex
32{
33 class STLtools;
34}
35
36namespace amrex::EB2 {
37
38class IndexSpace;
39template <typename G> class GShopLevel;
40template <typename G> class IndexSpaceImp;
41
42class Level
43{
44public:
45
47 bool isAllRegular () const noexcept { return m_allregular; }
49 bool isOK () const noexcept { return m_ok; }
60 void fillVolFrac (MultiFab& vfrac, const Geometry& geom) const;
64 void fillCentroid (MultiCutFab& centroid, const Geometry& geom) const;
66 void fillCentroid ( MultiFab& centroid, const Geometry& geom) const;
70 void fillBndryArea (MultiCutFab& bndryarea, const Geometry& geom) const;
72 void fillBndryArea ( MultiFab& bndryarea, const Geometry& geom) const;
76 void fillBndryCent (MultiCutFab& bndrycent, const Geometry& geom) const;
78 void fillBndryCent ( MultiFab& bndrycent, const Geometry& geom) const;
82 void fillBndryNorm (MultiCutFab& bndrynorm, const Geometry& geom) const;
84 void fillBndryNorm ( MultiFab& bndrynorm, const Geometry& geom) const;
88 void fillAreaFrac (Array<MultiCutFab*,AMREX_SPACEDIM> const& areafrac, const Geometry& geom) const;
90 void fillAreaFrac (Array< MultiFab*,AMREX_SPACEDIM> const& areafrac, const Geometry& geom) const;
94 void fillFaceCent (Array<MultiCutFab*,AMREX_SPACEDIM> const& facecent, const Geometry& geom) const;
96 void fillFaceCent (Array< MultiFab*,AMREX_SPACEDIM> const& facecent, const Geometry& geom) const;
100 void fillEdgeCent (Array<MultiCutFab*,AMREX_SPACEDIM> const& edgecent, const Geometry& geom) const;
102 void fillEdgeCent (Array< MultiFab*,AMREX_SPACEDIM> const& edgecent, const Geometry& geom) const;
106 void fillLevelSet (MultiFab& levelset, const Geometry& geom) const;
107
109 const BoxArray& boxArray () const noexcept { return m_grids; }
111 const DistributionMapping& DistributionMap () const noexcept { return m_dmap; }
112
114 Level (IndexSpace const* is, const Geometry& geom) : m_geom(geom), m_parent(is) {}
118 void prepareForCoarsening (const Level& rhs, int max_grid_size, IntVect const& ngrow);
119
121 const Geometry& Geom () const noexcept { return m_geom; }
123 IndexSpace const* getEBIndexSpace () const noexcept { return m_parent; }
124
126 IntVect const& nGrowVect () const noexcept { return m_ngrow; }
127
135 void write_to_chkpt_file (const std::string& fname, bool extend_domain_face, int max_grid_size) const;
136
138 bool hasEBInfo () const noexcept { return m_has_eb_info; }
140 void fillCutCellMask (iMultiFab& cutcellmask, const Geometry& geom) const;
141
145 void setShift (int direction, int ncells);
146
150 struct FCData {
160 // Covered region of the refined level, coarsened onto this level. The refined level
161 // is transient, so this has to be captured while it exists. It is not the same as
162 // Level::m_covered_grids: the refined level resolves the body more finely, so once
163 // coarsened its covered region is the larger of the two, and it is the one that
164 // matches where the face-centered data was computed.
165 //
166 // The fill*FC functions apply it before their ParallelCopy, and deliberately do not
167 // shrink it first. A face-centered cell straddles two cell-centered cells, so the
168 // ones on the covered/cut interface may themselves be cut; clearing them here and
169 // letting the copy write over them is correct, because the copy writes exactly where
170 // the face-centered data is valid - the nodal span of the cut boxes, which reaches
171 // the interface and no further. Ordering does the shrinking, so no box has to be
172 // trimmed and two adjacent covered boxes need no special treatment.
174 bool m_built = false;
175 };
176
185 [[nodiscard]] static IntVect fcEdgeType (int idim, int face_dir) noexcept {
186 IntVect edge_type{1};
187 if (idim != face_dir) { edge_type[idim] = 0; }
188 return edge_type;
189 }
190
192 [[nodiscard]] bool hasFCData (int face_dir) const noexcept {
193 return face_dir >= 0 && face_dir < AMREX_SPACEDIM &&
194 m_fc_data[face_dir] && m_fc_data[face_dir]->m_built;
195 }
196
198 [[nodiscard]] const FCData& getFCData (int face_dir) const {
199 AMREX_ASSERT(hasFCData(face_dir));
200 return *m_fc_data[face_dir];
201 }
202
204 void fillVolFracFC (MultiFab& vfrac, int face_dir, const Geometry& geom) const;
206 void fillAreaFracFC (Array< MultiFab*,AMREX_SPACEDIM> const& areafrac, int face_dir, const Geometry& geom) const;
207 void fillAreaFracFC (Array<MultiCutFab*,AMREX_SPACEDIM> const& areafrac, int face_dir, const Geometry& geom) const;
209 void fillCentroidFC (MultiFab& centroid, int face_dir, const Geometry& geom) const;
210 void fillCentroidFC (MultiCutFab& centroid, int face_dir, const Geometry& geom) const;
212 void fillBndryAreaFC (MultiFab& bndryarea, int face_dir, const Geometry& geom) const;
213 void fillBndryAreaFC (MultiCutFab& bndryarea, int face_dir, const Geometry& geom) const;
215 void fillBndryCentFC (MultiFab& bndrycent, int face_dir, const Geometry& geom) const;
216 void fillBndryCentFC (MultiCutFab& bndrycent, int face_dir, const Geometry& geom) const;
218 void fillBndryNormFC (MultiFab& bndrynorm, int face_dir, const Geometry& geom) const;
219 void fillBndryNormFC (MultiCutFab& bndrynorm, int face_dir, const Geometry& geom) const;
221 void fillEBCellFlagFC (FabArray<EBCellFlagFab>& cellflag, int face_dir, const Geometry& geom) const;
223 void fillFaceCentFC (Array< MultiFab*,AMREX_SPACEDIM> const& facecent, int face_dir, const Geometry& geom) const;
224 void fillFaceCentFC (Array<MultiCutFab*,AMREX_SPACEDIM> const& facecent, int face_dir, const Geometry& geom) const;
226 void fillEdgeCentFC (Array< MultiFab*,AMREX_SPACEDIM> const& edgecent, int face_dir, const Geometry& geom) const;
227 void fillEdgeCentFC (Array<MultiCutFab*,AMREX_SPACEDIM> const& edgecent, int face_dir, const Geometry& geom) const;
228
229// public: // for cuda
237 int coarsenFromFine (Level& fineLevel, bool fill_boundary);
239 void buildCellFlag ();
241 void buildCellFlagFC (int face_dir);
243 void buildCutCellMask (Level const& fine_level);
244
245protected:
246
253#if (AMREX_SPACEDIM == 3)
254 std::map<int,std::unique_ptr<MC::MCFab>> m_marching_cubes;
255#endif
268 bool m_allregular = false;
269 bool m_ok = false;
270 bool m_has_eb_info = true;
273
276
277private:
278 template <typename G> friend class GShopLevel;
279 // So that it can update m_parent when a level changes owner.
280 template <typename G> friend class IndexSpaceImp;
281
282 // Need this function to work around a gcc bug.
283 void setRegularLevel () {
284 m_allregular = true;
285 m_ok = true;
286 m_has_eb_info = false;
287 }
288};
289
290template <typename G>
292 : public Level
293{
294public:
298 GShopLevel (IndexSpace const* is, G const& gshop, const Geometry& geom, int max_grid_size,
299 int ngrow, bool extend_domain_face, int num_crse_opt);
303 GShopLevel (IndexSpace const* is, int ilev, int max_grid_size, int ngrow,
304 const Geometry& geom, GShopLevel<G>& fineLevel);
306 GShopLevel (IndexSpace const* is, const Geometry& geom);
308 void define_fine (G const& gshop, const Geometry& geom,
309 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt);
310
311#if (AMREX_SPACEDIM == 3)
315 template <typename GS = G>
316 requires (std::same_as<GS,STLtools>)
317 void define_fine_mvmc (GS const& gshop, const Geometry& geom,
318 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt);
319#endif
320
324 static GShopLevel<G>
325 makeAllRegular(IndexSpace const* is, const Geometry& geom)
326 {
327 GShopLevel<G> r(is, geom);
328 r.setRegularLevel();
329 return r;
330 }
331
333 void buildFCData (G const& gshop, int face_dir, int max_grid_size);
334
335private:
336
337 Box m_bounding_box;
338
339 void prepare_grids (G const& gshop, Geometry const& geom, int max_grid_size, int ngrow,
340 bool extend_domain_face, int num_crse_opt);
341
342};
343
344template <typename G>
346 : Level(is, geom)
347{}
348
349template <typename G>
350GShopLevel<G>::GShopLevel (IndexSpace const* is, G const& gshop, const Geometry& geom,
351 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
352 : Level(is, geom)
353{
354 if (std::is_same_v<typename G::FunctionType, AllRegularIF>) {
355 m_allregular = true;
356 m_ok = true;
357 return;
358 }
359
360 define_fine(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
361}
362
363template <typename G>
364void
365GShopLevel<G>::prepare_grids (G const& gshop, const Geometry& geom,
366 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
367{
368 if (amrex::Verbose() > 0 && extend_domain_face == false) {
369 amrex::Print() << "AMReX WARNING: extend_domain_face=false is not recommended!\n";
370 }
371
372 BL_PROFILE("EB2::GShopLevel()-prepare_grids");
373
374 // make sure ngrow is multiple of 16
375 m_ngrow = IntVect{static_cast<int>(std::ceil(ngrow/16.)) * 16};
376
377 Box const& domain = geom.Domain();
378 Box domain_grown = domain;
379 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
380 if (geom.isPeriodic(idim)) {
381 m_ngrow[idim] = 0;
382 } else {
383 m_ngrow[idim] = std::min(m_ngrow[idim], domain_grown.length(idim));
384 }
385 }
386 domain_grown.grow(m_ngrow);
387 m_bounding_box = (extend_domain_face) ? domain : domain_grown;
388 m_bounding_box.surroundingNodes();
389
390 BoxList cut_boxes;
391 BoxList covered_boxes;
392
393 const int nprocs = ParallelContext::NProcsSub();
394 const int iproc = ParallelContext::MyProcSub();
395
396 num_crse_opt = std::max(0,std::min(8,num_crse_opt));
397 for (int clev = num_crse_opt; clev >= 0; --clev) {
398 IntVect crse_ratio(1 << clev);
399 if (domain.coarsenable(crse_ratio)) {
400 Box const& crse_bounding_box = amrex::coarsen(m_bounding_box, crse_ratio);
401 Geometry const& crse_geom = amrex::coarsen(geom, crse_ratio);
402 BoxList test_boxes;
403 if (cut_boxes.isEmpty()) {
404 covered_boxes.clear();
405 test_boxes = BoxList(crse_geom.Domain());
406 test_boxes.maxSize(max_grid_size);
407 } else {
408 test_boxes.swap(cut_boxes);
409 test_boxes.coarsen(crse_ratio);
410 test_boxes.maxSize(max_grid_size);
411 }
412
413 const Long nboxes = test_boxes.size();
414 const auto& boxes = test_boxes.data();
415 for (Long i = iproc; i < nboxes; i += nprocs) {
416 const Box& vbx = boxes[i];
417 const Box& gbx = amrex::surroundingNodes(amrex::grow(vbx,1));
418 auto box_type = gshop.getBoxType(gbx&crse_bounding_box,crse_geom,RunOn::Gpu);
419 if (box_type == gshop.allcovered) {
420 covered_boxes.push_back(amrex::refine(vbx, crse_ratio));
421 } else if (box_type == gshop.mixedcells) {
422 cut_boxes.push_back(amrex::refine(vbx, crse_ratio));
423 }
424 }
425
426 amrex::AllGatherBoxes(cut_boxes.data());
427 }
428 }
429
430 amrex::AllGatherBoxes(covered_boxes.data());
431
432 if (m_ngrow != 0) {
433 auto grow_at_domain_boundary = [&] (BoxList& bl)
434 {
435 const IntVect& domlo = domain.smallEnd();
436 const IntVect& domhi = domain.bigEnd();
437 for (auto& b : bl) {
438 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
439 if (m_ngrow[idim] != 0) {
440 if (b.smallEnd(idim) == domlo[idim]) {
441 b.growLo(idim,m_ngrow[idim]);
442 }
443 if (b.bigEnd(idim) == domhi[idim]) {
444 b.growHi(idim,m_ngrow[idim]);
445 }
446 }
447 }
448 }
449 };
450 grow_at_domain_boundary(covered_boxes);
451 grow_at_domain_boundary(cut_boxes);
452 }
453
454 if ( cut_boxes.isEmpty() &&
455 !covered_boxes.isEmpty())
456 {
457 amrex::Abort("AMReX_EB2_Level.H: Domain is completely covered");
458 }
459
460 if (!covered_boxes.isEmpty()) {
461 if (num_crse_opt > 2) { // don't want the box too big
462 covered_boxes.maxSize(max_grid_size*4);
463 }
464 m_covered_grids = BoxArray(std::move(covered_boxes));
465 }
466
467 if (cut_boxes.isEmpty()) {
468 m_grids = BoxArray();
469 m_dmap = DistributionMapping();
470 m_allregular = true;
471 m_ok = true;
472 } else {
473 m_grids = BoxArray(std::move(cut_boxes));
474 m_dmap = DistributionMapping(m_grids);
475 }
476}
477
478template <typename G>
479void
480GShopLevel<G>::define_fine (G const& gshop, const Geometry& geom,
481 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
482{
483 BL_PROFILE("EB2::GShopLevel()-fine");
484
485#ifdef AMREX_USE_FLOAT
486 Real small_volfrac = 1.e-5_rt;
487#else
488 Real small_volfrac = 1.e-14;
489#endif
490 bool cover_multiple_cuts = false;
491 int maxiter = 32;
492 {
493 ParmParse pp("eb2");
494 pp.queryAdd("small_volfrac", small_volfrac);
495 pp.queryAdd("cover_multiple_cuts", cover_multiple_cuts);
496 pp.queryAdd("maxiter", maxiter);
497 }
498 maxiter = std::min(100000, maxiter);
499
500 prepare_grids(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
501
502 if (m_allregular) { return; }
503
504 m_mgf.define(m_grids, m_dmap);
505 const int ng = GFab::ng;
506 MFInfo mf_info;
507 mf_info.SetTag("EB2::Level");
508 m_cellflag.define(m_grids, m_dmap, 1, ng, mf_info);
509 m_volfrac.define(m_grids, m_dmap, 1, ng, mf_info);
510 m_centroid.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
511 m_bndryarea.define(m_grids, m_dmap, 1, ng, mf_info);
512 m_bndrycent.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
513 m_bndrynorm.define(m_grids, m_dmap, AMREX_SPACEDIM, ng, mf_info);
514 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
515 m_areafrac[idim].define(amrex::convert(m_grids, IntVect::TheDimensionVector(idim)),
516 m_dmap, 1, ng, mf_info);
517 m_facecent[idim].define(amrex::convert(m_grids, IntVect::TheDimensionVector(idim)),
518 m_dmap, AMREX_SPACEDIM-1, ng, mf_info);
519 IntVect edge_type{1}; edge_type[idim] = 0;
520 m_edgecent[idim].define(amrex::convert(m_grids, edge_type), m_dmap, 1, ng, mf_info);
521 }
522
523 const auto dx = geom.CellSizeArray();
524 const auto problo = geom.ProbLoArray();
525
526 auto bounding_box = m_bounding_box;
527 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
528 if (!extend_domain_face || geom.isPeriodic(idim)) {
529 bounding_box.grow(idim,GFab::ng);
530 }
531 }
532
533 RunOn gshop_run_on = (Gpu::inLaunchRegion() && gshop.isGPUable())
535
536 bool hybrid = Gpu::inLaunchRegion() && (gshop_run_on == RunOn::Cpu);
537 amrex::ignore_unused(hybrid);
538
539 int iter = 0;
540 for (; iter < maxiter; ++iter)
541 {
542 int nsmallcells = 0;
543 int nmulticuts = 0;
544#ifdef AMREX_USE_OMP
545#pragma omp parallel if (Gpu::notInLaunchRegion()) reduction(+:nsmallcells,nmulticuts)
546#endif
547 {
548#if (AMREX_SPACEDIM == 3)
549 Array<BaseFab<Real>, AMREX_SPACEDIM> M2;
550 EBCellFlagFab cellflagtmp;
551#endif
552 for (MFIter mfi(m_mgf); mfi.isValid(); ++mfi)
553 {
554 auto& gfab = m_mgf[mfi];
555 const Box& vbx = gfab.validbox();
556
557 auto& levelset = gfab.getLevelSet();
558 if (iter == 0) {
559 gshop.fillFab(levelset, geom, gshop_run_on, bounding_box);
560#ifdef AMREX_USE_GPU
561 if (hybrid) {
562 levelset.prefetchToDevice();
563 }
564#endif
565 }
566
567 auto& cellflag = m_cellflag[mfi];
568
569 gfab.buildTypes(cellflag);
570
571 Array4<Real const> const& clst = levelset.const_array();
572 Array4<Real > const& lst = levelset.array();
573 Array4<EBCellFlag> const& cfg = m_cellflag.array(mfi);
574 Array4<Real> const& vfr = m_volfrac.array(mfi);
575 Array4<Real> const& ctr = m_centroid.array(mfi);
576 Array4<Real> const& bar = m_bndryarea.array(mfi);
577 Array4<Real> const& bct = m_bndrycent.array(mfi);
578 Array4<Real> const& bnm = m_bndrynorm.array(mfi);
579 AMREX_D_TERM(Array4<Real> const& apx = m_areafrac[0].array(mfi);,
580 Array4<Real> const& apy = m_areafrac[1].array(mfi);,
581 Array4<Real> const& apz = m_areafrac[2].array(mfi););
582 AMREX_D_TERM(Array4<Real> const& fcx = m_facecent[0].array(mfi);,
583 Array4<Real> const& fcy = m_facecent[1].array(mfi);,
584 Array4<Real> const& fcz = m_facecent[2].array(mfi););
585
586 auto& facetype = gfab.getFaceType();
587 AMREX_D_TERM(Array4<Type_t> const& ftx = facetype[0].array();,
588 Array4<Type_t> const& fty = facetype[1].array();,
589 Array4<Type_t> const& ftz = facetype[2].array(););
590
591 int nmc = 0;
592 int nsm = 0;
593
594#if (AMREX_SPACEDIM == 3)
595 auto& edgetype = gfab.getEdgeType();
596 Array4<Type_t const> const& xdg = edgetype[0].const_array();
597 Array4<Type_t const> const& ydg = edgetype[1].const_array();
598 Array4<Type_t const> const& zdg = edgetype[2].const_array();
599
600 Array4<Real> const& xip = m_edgecent[0].array(mfi);
601 Array4<Real> const& yip = m_edgecent[1].array(mfi);
602 Array4<Real> const& zip = m_edgecent[2].array(mfi);
603
604 if (iter == 0) {
605#ifdef AMREX_USE_GPU
606 if (hybrid) {
608 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
609 edgetype[idim].prefetchToHost();
610 m_edgecent[idim][mfi].prefetchToHost();
611 }
612 }
613#endif
614
615 gshop.getIntercept({xip,yip,zip}, {xdg,ydg,zdg}, clst,
616 geom, gshop_run_on, bounding_box);
617
618#ifdef AMREX_USE_GPU
619 if (hybrid) {
620 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
621 edgetype[idim].prefetchToDevice();
622 m_edgecent[idim][mfi].prefetchToDevice();
623 }
624 }
625#endif
626 } else {
627 gshop.updateIntercept({xip,yip,zip}, {xdg,ydg,zdg}, clst, geom);
628 }
629
630 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
631 const Box& b = facetype[idim].box();
632 M2[idim].resize(b,3);
633 }
634 Array4<Real> const& xm2 = M2[0].array();
635 Array4<Real> const& ym2 = M2[1].array();
636 Array4<Real> const& zm2 = M2[2].array();
637
638 nmc = build_faces(vbx, cfg, ftx, fty, ftz, xdg, ydg, zdg, lst,
639 xip, yip, zip, apx, apy, apz, fcx, fcy, fcz,
640 xm2, ym2, zm2, dx, problo, cover_multiple_cuts);
641
642 cellflagtmp.resize(m_cellflag[mfi].box());
643 Elixir cellflagtmp_eli = cellflagtmp.elixir();
644 Array4<EBCellFlag> const& cfgtmp = cellflagtmp.array();
645
646 build_cells(vbx, cfg, ftx, fty, ftz, apx, apy, apz,
647 fcx, fcy, fcz, xm2, ym2, zm2, dx, vfr, ctr,
648 bar, bct, bnm, cfgtmp, lst,
649 small_volfrac, geom, extend_domain_face, cover_multiple_cuts,
650 nsm, nmc);
651
652 // Because it is used in a synchronous reduction kernel in
653 // build_cells, we do not need to worry about M2's lifetime.
654 // But we still need to use Elixir to extend the life of
655 // cellflagtmp.
656
657#elif (AMREX_SPACEDIM == 2)
658 Array4<Real> const& xip = m_edgecent[0].array(mfi);
659 Array4<Real> const& yip = m_edgecent[1].array(mfi);
660
661 if (iter == 0) {
662#ifdef AMREX_USE_GPU
663 if (hybrid) {
665 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
666 facetype[idim].prefetchToHost();
667 m_edgecent[idim][mfi].prefetchToHost();
668 }
669 }
670#endif
671
672 // yes, factype[1] and then [0]
673 gshop.getIntercept({xip,yip},
674 {facetype[1].const_array(), facetype[0].const_array()},
675 clst, geom, gshop_run_on, bounding_box);
676
677#ifdef AMREX_USE_GPU
678 if (hybrid) {
679 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
680 facetype[idim].prefetchToDevice();
681 m_edgecent[idim][mfi].prefetchToDevice();
682 }
683 }
684#endif
685 } else {
686 gshop.updateIntercept({xip,yip},
687 {facetype[1].const_array(), facetype[0].const_array()},
688 clst, geom);
689 }
690
691 nmc = build_faces(vbx, cfg, ftx, fty, lst, xip, yip, apx, apy, fcx, fcy,
692 dx, problo, cover_multiple_cuts, nsm);
693
694 build_cells(vbx, cfg, ftx, fty, apx, apy, dx, vfr, ctr,
695 bar, bct, bnm, lst, small_volfrac, geom, extend_domain_face,
696 nsm, nmc);
697#endif
698 nsmallcells += nsm;
699 nmulticuts += nmc;
700 }
701 }
702
703 ParallelAllReduce::Sum<int>({nsmallcells,nmulticuts}, ParallelContext::CommunicatorSub());
704 if (nsmallcells == 0 && nmulticuts == 0) {
705 break;
706 } else {
707 auto ls = m_mgf.getLevelSet();
708 // This is an alias MultiFab, therefore FillBoundary on it is fine.
709 ls.FillBoundary(geom.periodicity());
710 if (amrex::Verbose() > 0) {
711 if (nsmallcells) {
712 amrex::Print() << "AMReX EB: Iter. " << iter+1 << " fixed " << nsmallcells
713 << " small cells" << '\n';
714 }
715 if (nmulticuts) {
716 amrex::Print() << "AMReX EB: Iter. " << iter+1 << " fixed " << nmulticuts
717 << " multicuts" << '\n';
718 }
719 }
720 }
721 }
722
723 AMREX_ALWAYS_ASSERT_WITH_MESSAGE(iter < maxiter, "EB: failed to fix small cells");
724
725#ifdef AMREX_USE_OMP
726#pragma omp parallel if (Gpu::notInLaunchRegion())
727#endif
728 for (MFIter mfi(m_mgf); mfi.isValid(); ++mfi)
729 {
730 auto& gfab = m_mgf[mfi];
731 auto const& levelset = gfab.getLevelSet();
732 Array4<Real const> const& clst = levelset.const_array();
733
734 AMREX_D_TERM(Array4<Real> const& xip = m_edgecent[0].array(mfi);,
735 Array4<Real> const& yip = m_edgecent[1].array(mfi);,
736 Array4<Real> const& zip = m_edgecent[2].array(mfi);)
737#if (AMREX_SPACEDIM == 3)
738 auto const& edgetype = gfab.getEdgeType();
739 Array4<Type_t const> const& xdg = edgetype[0].const_array();
740 Array4<Type_t const> const& ydg = edgetype[1].const_array();
741 Array4<Type_t const> const& zdg = edgetype[2].const_array();
742
743 intercept_to_edge_centroid(xip, yip, zip, xdg, ydg, zdg, clst, dx, problo);
744
745#elif (AMREX_SPACEDIM == 2)
746 auto& facetype = gfab.getFaceType();
747 Array4<Type_t const> const& ftx = facetype[0].const_array();
748 Array4<Type_t const> const& fty = facetype[1].const_array();
749 // fty then ftx
750 intercept_to_edge_centroid(xip, yip, fty, ftx, clst, dx, problo);
751#endif
752 }
753
754 m_levelset = m_mgf.getLevelSet();
755
756 m_ok = true;
757}
758
759#if (AMREX_SPACEDIM == 3)
760template <typename G>
761template <typename GS>
762requires (std::same_as<GS,STLtools>)
763void
764GShopLevel<G>::define_fine_mvmc (GS const& gshop, const Geometry& geom,
765 int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
766{
767 BL_PROFILE("EB2::GShopLevel()-fine-mvmc");
768
769 prepare_grids(gshop, geom, max_grid_size, ngrow, extend_domain_face, num_crse_opt);
770
771 if (m_allregular) { return; }
772
774
775 m_sdf.define(amrex::convert(m_grids,IntVect(1)), m_dmap, 1, IntVect(1));
776 gshop.fillSignedDistance(m_sdf, m_sdf.nGrowVect(), geom);
777
778 for (MFIter mfi(m_sdf,MFItInfo().DisableDeviceSync()); mfi.isValid(); ++mfi) {
779 m_marching_cubes[mfi.index()] = std::make_unique<MC::MCFab>();
780 }
781
782 MFItInfo info{};
783
784#if defined(AMREX_USE_OMP) && !defined(AMREX_USE_GPU)
785 info.SetDynamic(true);
786#pragma omp parallel
787#endif
788 for (MFIter mfi(m_sdf,info); mfi.isValid(); ++mfi) {
789 amrex::MC::marching_cubes(geom, m_sdf[mfi], *m_marching_cubes[mfi.index()]);
790 }
791
792 MC::write_stl("test.stl", m_marching_cubes);
793
794 m_ok = false; // xxxxx TODO
795}
796#endif
797
798template <typename G>
799GShopLevel<G>::GShopLevel (IndexSpace const* is, int /*ilev*/, int max_grid_size, int /*ngrow*/,
800 const Geometry& geom, GShopLevel<G>& fineLevel)
801 : Level(is, geom)
802{
803 if (fineLevel.isAllRegular()) {
804 m_allregular = true;
805 m_ok = true;
806 return;
807 }
808
809 BL_PROFILE("EB2::GShopLevel()-coarse");
810
811 const BoxArray& fine_grids = fineLevel.m_grids;
812 const BoxArray& fine_covered_grids = fineLevel.m_covered_grids;
813
814 const int coarse_ratio = 2;
815 const int min_width = 8;
816 bool coarsenable = fine_grids.coarsenable(coarse_ratio, min_width)
817 && (fine_covered_grids.empty() || fine_covered_grids.coarsenable(coarse_ratio));
818
819 m_ngrow = amrex::coarsen(fineLevel.m_ngrow,2);
820 if (amrex::scale(m_ngrow,2) != fineLevel.m_ngrow) {
822 }
823
824 if (coarsenable)
825 {
826 int ierr = coarsenFromFine(fineLevel, true);
827 m_ok = (ierr == 0);
828 }
829 else
830 {
831 Level fine_level_2(is, fineLevel.m_geom);
832 fine_level_2.prepareForCoarsening(fineLevel, max_grid_size, amrex::scale(m_ngrow,2));
833 int ierr = coarsenFromFine(fine_level_2, false);
834 m_ok = (ierr == 0);
835 }
836}
837
838template <typename G>
839void
840GShopLevel<G>::buildFCData (G const& gshop, int face_dir, int max_grid_size)
841{
842 BL_PROFILE("EB2::GShopLevel::buildFCData");
843
844 AMREX_ASSERT(face_dir >= 0 && face_dir < AMREX_SPACEDIM);
845
846 // Allocate FC data structure if needed
847 if (!m_fc_data[face_dir]) {
848 m_fc_data[face_dir] = std::make_unique<FCData>();
849 }
850
851 if (m_fc_data[face_dir]->m_built) { return; }
852
853 if (m_allregular) {
854 // An all-regular level carries no MultiFabs, and there is nothing to coarsen. Mark the
855 // face-centered data built anyway so that hasFCData() is true and a face-centered factory
856 // can be constructed: the fill*FC functions fill the regular values for such a level
857 // without reading anything, exactly as their cell-centered counterparts do.
858 m_fc_data[face_dir]->m_built = true;
859 return;
860 }
861
862 // **********************************************************************
863 // STEP1: Get or build the one-level-finer CC data
864 // **********************************************************************
865
866 Level const* refined_cc_level = nullptr;
867 std::unique_ptr<GShopLevel<G>> transient_level;
868
869 // Always create transient refined CC level to avoid grid mismatch
870 // with application-defined BoxArrays (ERF EB-FC development)
871 Geometry refined_geom = amrex::refine(m_geom, 2);
872 transient_level = std::make_unique<GShopLevel<G>>(
873 m_parent, gshop, refined_geom,
874 max_grid_size, m_ngrow[0] > 0 ? m_ngrow[0] : 4, true, 0); // xxxxx TODO: should we use IndexSpace's member variables extend_domain_face and num_crse_opt?
875
876 if (!transient_level->isOK()) {
877 amrex::Abort("GShopLevel::buildFCData: failed to build refined level");
878 }
879 refined_cc_level = transient_level.get();
880
881 // **********************************************************************
882 // STEP2: Check multi-cut error
883 // **********************************************************************
884
885 // Everything below is local. m_grids, m_dmap and m_levelset describe the cell-centered
886 // level that the IndexSpace hands to every EBFArrayBoxFactory built from it, so building
887 // face-centered data must not touch them: overwriting m_grids would leave Level::boxArray()
888 // describing a different region than the cell-centered data it belongs to, and redefining
889 // m_levelset on those grids would silently change the level set that cell-centered
890 // factories created afterwards report.
891 //
892 // The grids are coarsened from the refined level rather than taken from m_grids so that the
893 // DistributionMapping stays compatible with the refined data we read below.
894 const BoxArray fc_base_grids = amrex::coarsen(refined_cc_level->m_grids, 2);
895 if (!refined_cc_level->m_covered_grids.empty()) {
896 m_fc_data[face_dir]->m_covered_grids_fc =
897 amrex::coarsen(refined_cc_level->m_covered_grids, 2).simplified();
898 }
899 const DistributionMapping fc_dmap = refined_cc_level->m_dmap;
900
901 // Face-centered data provides no level set, so nothing here coarsens one. The multi-cut check
902 // below only needs an MFIter over the coarse boxes, walking alongside the refined level set
903 // that shares this DistributionMapping, so this MultiFab is declared without data.
904 auto const& f_levelset = refined_cc_level->m_levelset;
905 MultiFab crse_boxes(amrex::convert(fc_base_grids,IntVect::TheNodeVector()), fc_dmap, 1, 0,
906 MFInfo{}.SetAlloc(false));
907
908 int mvmc_error = 0;
909
911 {
912#ifdef AMREX_USE_OMP
913#pragma omp parallel reduction(max:mvmc_error)
914#endif
915 for (MFIter mfi(crse_boxes,true); mfi.isValid(); ++mfi)
916 {
917 const Box& ccbx = mfi.tilebox(IntVect::TheCellVector());
918 auto const& fine = f_levelset.const_array(mfi);
919
920 int tile_error = 0;
921 amrex::LoopOnCpu(ccbx,
922 [=,&tile_error] (int i, int j, int k) noexcept
923 {
924 int ierror = check_mvmc(i,j,k,fine);
925 tile_error = std::max(tile_error,ierror);
926 });
927
928 mvmc_error = std::max(mvmc_error, tile_error);
929 }
930 }
931 else
932 {
933 ReduceOps<ReduceOpMax> reduce_op;
934 ReduceData<int> reduce_data(reduce_op);
935 using ReduceTuple = typename decltype(reduce_data)::Type;
936
937 for (MFIter mfi(crse_boxes); mfi.isValid(); ++mfi)
938 {
939 const Box& ndbx = mfi.validbox();
940 const Box& ccbx = amrex::enclosedCells(ndbx);
941 auto const& fine = f_levelset.const_array(mfi);
942 reduce_op.eval(ndbx, reduce_data,
943 [=] AMREX_GPU_DEVICE (int i, int j, int k) -> ReduceTuple
944 {
945 int ierror;
946 if (ccbx.contains(IntVect{AMREX_D_DECL(i,j,k)})) {
947 ierror = check_mvmc(i,j,k,fine);
948 } else {
949 ierror = 0;
950 }
951 return {ierror};
952 });
953 }
954 ReduceTuple rv = reduce_data.value(reduce_op);
955 mvmc_error = amrex::max(0, amrex::get<0>(rv));
956 }
957
958 {
959 bool b = mvmc_error;
961 mvmc_error = b;
962 }
963 if (mvmc_error) {
964 amrex::Abort("GShopLevel::buildFCData: MVMC error");
965 }
966
967 // **********************************************************************
968 // STEP3: Define FC data structures
969 // **********************************************************************
970
971 // FC MultiFabs
972 auto& fc_cellflag = m_fc_data[face_dir]->m_cellflag_fc;
973 MultiFab& fc_volfrac = m_fc_data[face_dir]->m_volfrac_fc;
974 MultiFab& fc_centroid = m_fc_data[face_dir]->m_centroid_fc;
975 MultiFab& fc_bndryarea = m_fc_data[face_dir]->m_bndryarea_fc;
976 MultiFab& fc_bndrycent = m_fc_data[face_dir]->m_bndrycent_fc;
977 MultiFab& fc_bndrynorm = m_fc_data[face_dir]->m_bndrynorm_fc;
978 auto& fc_areafrac = m_fc_data[face_dir]->m_areafrac_fc;
979 auto& fc_facecent = m_fc_data[face_dir]->m_facecent_fc;
980 auto& fc_edgecent = m_fc_data[face_dir]->m_edgecent_fc;
981
982 // Index types of the face-centered data. One rule covers all of it: face_dir is nodal,
983 // because that is how a box spans staggered cells - there is one more of them than there
984 // are cell-centered cells, and nodal is how AMReX says so. The remaining directions carry
985 // whatever type the cell-centered path would give.
986 //
987 // cellflag / volfrac / centroid / bndry* nodal in face_dir alone, i.e. (N,C,C) for
988 // face_dir=0, (C,N,C) for 1, (C,C,N) for 2.
989 // These are the staggered cells themselves.
990 // areafrac[idim] / facecent[idim] the same, for every idim.
991 // edgecent[idim] cell in idim and nodal elsewhere, as in the
992 // cell-centered path, with face_dir forced nodal.
993 //
994 // areafrac and facecent are kept uniform across idim on purpose. The cell-centered path
995 // gives each direction its own nodal index type (areafrac[0] is (N,C,C), areafrac[1] is
996 // (C,N,C), ...); carrying that over to a grid already staggered in face_dir would put three
997 // different index types in play at once for no real gain. Here all three are simply indexed
998 // by the staggered cell they belong to, and areafrac[idim](i,j,k) is that cell's low face in
999 // direction idim, with areafrac[idim](i+1,j,k) its high one.
1000 //
1001 // Indexing by the staggered cell is why face_dir has to be nodal even here. The valid box
1002 // then holds every staggered cell the box owns, which is what coarsen_from_fine below fills
1003 // and what the fill*FC ParallelCopy can hand out. A cell-typed face_dir would stop one short
1004 // and leave the last staggered cell of every box in the ghost region, where the copy cannot
1005 // reach it: the covered-region mask in fill*FC would clear that plane with nothing able to
1006 // write it back.
1007
1008 // Define FC MultiFabs
1009 BoxArray fc_grids = amrex::convert(fc_base_grids, IntVect::TheDimensionVector(face_dir));
1010 const int ng = 3;
1011 fc_cellflag.define(fc_grids, fc_dmap, 1, ng);
1012 fc_volfrac.define(fc_grids, fc_dmap, 1, ng);
1013 fc_centroid.define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1014 fc_bndryarea.define(fc_grids, fc_dmap, 1, ng);
1015 fc_bndrycent.define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1016 fc_bndrynorm.define(fc_grids, fc_dmap, AMREX_SPACEDIM, ng);
1017
1018 for (int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
1019 fc_areafrac[idim].define(fc_grids, fc_dmap, 1, ng);
1020 fc_facecent[idim].define(fc_grids, fc_dmap, AMREX_SPACEDIM-1, ng);
1021 fc_edgecent[idim].define(amrex::convert(fc_base_grids, Level::fcEdgeType(idim,face_dir)),
1022 fc_dmap, 1, ng);
1023 }
1024
1025 // **********************************************************************
1026 // STEP4: Coarsen from refined CC to FC with staggered averaging
1027 // **********************************************************************
1028
1029 // Fine CC MultiFabs
1030 auto const& cc_cellflag = refined_cc_level->m_cellflag;
1031 MultiFab const& cc_volfrac = refined_cc_level->m_volfrac;
1032 MultiFab const& cc_centroid = refined_cc_level->m_centroid;
1033 MultiFab const& cc_bndryarea = refined_cc_level->m_bndryarea;
1034 MultiFab const& cc_bndrycent = refined_cc_level->m_bndrycent;
1035 MultiFab const& cc_bndrynorm = refined_cc_level->m_bndrynorm;
1036 auto const& cc_areafrac = refined_cc_level->m_areafrac;
1037 auto const& cc_facecent = refined_cc_level->m_facecent;
1038 auto const& cc_edgecent = refined_cc_level->m_edgecent;
1039
1040 int error = 0;
1041
1043 {
1044#ifdef AMREX_USE_OMP
1045#pragma omp parallel reduction(max:error)
1046#endif
1047 for (MFIter mfi(fc_volfrac,true); mfi.isValid(); ++mfi)
1048 {
1049 // Coarse FC arrays
1050 auto const& cvol = fc_volfrac.array(mfi);
1051 auto const& ccent = fc_centroid.array(mfi);
1052 auto const& cba = fc_bndryarea.array(mfi);
1053 auto const& cbc = fc_bndrycent.array(mfi);
1054 auto const& cbn = fc_bndrynorm.array(mfi);
1055 AMREX_D_TERM(auto const& capx = fc_areafrac[0].array(mfi);,
1056 auto const& capy = fc_areafrac[1].array(mfi);,
1057 auto const& capz = fc_areafrac[2].array(mfi););
1058 AMREX_D_TERM(auto const& cfcx = fc_facecent[0].array(mfi);,
1059 auto const& cfcy = fc_facecent[1].array(mfi);,
1060 auto const& cfcz = fc_facecent[2].array(mfi););
1061 AMREX_D_TERM(auto const& cecx = fc_edgecent[0].array(mfi);,
1062 auto const& cecy = fc_edgecent[1].array(mfi);,
1063 auto const& cecz = fc_edgecent[2].array(mfi););
1064 auto const& cflag = fc_cellflag.array(mfi);
1065
1066 // Fine CC arrays
1067 auto const& fvol = cc_volfrac.const_array(mfi);
1068 auto const& fcent = cc_centroid.const_array(mfi);
1069 auto const& fba = cc_bndryarea.const_array(mfi);
1070 auto const& fbc = cc_bndrycent.const_array(mfi);
1071 auto const& fbn = cc_bndrynorm.const_array(mfi);
1072 AMREX_D_TERM(auto const& fapx = cc_areafrac[0].const_array(mfi);,
1073 auto const& fapy = cc_areafrac[1].const_array(mfi);,
1074 auto const& fapz = cc_areafrac[2].const_array(mfi););
1075 AMREX_D_TERM(auto const& ffcx = cc_facecent[0].const_array(mfi);,
1076 auto const& ffcy = cc_facecent[1].const_array(mfi);,
1077 auto const& ffcz = cc_facecent[2].const_array(mfi););
1078 AMREX_D_TERM(auto const& fecx = cc_edgecent[0].const_array(mfi);,
1079 auto const& fecy = cc_edgecent[1].const_array(mfi);,
1080 auto const& fecz = cc_edgecent[2].const_array(mfi););
1081 auto const& fflag = cc_cellflag.const_array(mfi);
1082
1083 Box const& bx = mfi.validbox();
1084 Box const& ndgbx = mfi.grownnodaltilebox(-1,2);
1085
1086 int tile_error = 0;
1087 amrex::LoopOnCpu(ndgbx,
1088 [=,&tile_error] (int i, int j, int k) noexcept
1089 {
1091 int ierr = coarsen_from_fine(AMREX_D_DECL(i,j,k), bx, 2,
1092 cvol,ccent,cba,cbc,cbn,
1093 AMREX_D_DECL(capx,capy,capz),
1094 AMREX_D_DECL(cfcx,cfcy,cfcz),
1095 AMREX_D_DECL(cecx,cecy,cecz),
1096 cflag,fvol,fcent,fba,fbc,fbn,
1097 AMREX_D_DECL(fapx,fapy,fapz),
1098 AMREX_D_DECL(ffcx,ffcy,ffcz),
1099 AMREX_D_DECL(fecx,fecy,fecz),
1100 fflag, face_dir);
1101 tile_error = std::max(tile_error,ierr);
1102 });
1103
1104 error = std::max(error,tile_error);
1105 }
1106 }
1107 else
1108 {
1109 ReduceOps<ReduceOpMax> reduce_op;
1110 ReduceData<int> reduce_data(reduce_op);
1111 using ReduceTuple = typename decltype(reduce_data)::Type;
1112
1113 for (MFIter mfi(fc_volfrac); mfi.isValid(); ++mfi)
1114 {
1115 auto const& cvol = fc_volfrac.array(mfi);
1116 auto const& ccent = fc_centroid.array(mfi);
1117 auto const& cba = fc_bndryarea.array(mfi);
1118 auto const& cbc = fc_bndrycent.array(mfi);
1119 auto const& cbn = fc_bndrynorm.array(mfi);
1120 AMREX_D_TERM(auto const& capx = fc_areafrac[0].array(mfi);,
1121 auto const& capy = fc_areafrac[1].array(mfi);,
1122 auto const& capz = fc_areafrac[2].array(mfi););
1123 AMREX_D_TERM(auto const& cfcx = fc_facecent[0].array(mfi);,
1124 auto const& cfcy = fc_facecent[1].array(mfi);,
1125 auto const& cfcz = fc_facecent[2].array(mfi););
1126 // We use MultiArray4 instead of Array4 here
1127 // to be below the 2 kB kernel parameter limit for SYCL
1128 AMREX_D_TERM(auto const& cecx = fc_edgecent[0].arrays();,
1129 auto const& cecy = fc_edgecent[1].arrays();,
1130 auto const& cecz = fc_edgecent[2].arrays(););
1131 auto const& cflag = fc_cellflag.array(mfi);
1132
1133 auto const& fvol = cc_volfrac.const_array(mfi);
1134 auto const& fcent = cc_centroid.const_array(mfi);
1135 auto const& fba = cc_bndryarea.const_array(mfi);
1136 auto const& fbc = cc_bndrycent.const_array(mfi);
1137 auto const& fbn = cc_bndrynorm.const_array(mfi);
1138 AMREX_D_TERM(auto const& fapx = cc_areafrac[0].const_array(mfi);,
1139 auto const& fapy = cc_areafrac[1].const_array(mfi);,
1140 auto const& fapz = cc_areafrac[2].const_array(mfi););
1141 AMREX_D_TERM(auto const& ffcx = cc_facecent[0].const_array(mfi);,
1142 auto const& ffcy = cc_facecent[1].const_array(mfi);,
1143 auto const& ffcz = cc_facecent[2].const_array(mfi););
1144 AMREX_D_TERM(auto const& fecx = cc_edgecent[0].const_arrays();,
1145 auto const& fecy = cc_edgecent[1].const_arrays();,
1146 auto const& fecz = cc_edgecent[2].const_arrays(););
1147 auto const& fflag = cc_cellflag.const_array(mfi);
1148
1149 const int lidx = mfi.LocalIndex();
1150 Box const& bx = mfi.validbox();
1151 Box const& gbx = amrex::grow(bx,2);
1152 Box const& ndgbx = amrex::surroundingNodes(gbx);
1153
1154 reduce_op.eval(ndgbx, reduce_data,
1155 [=] AMREX_GPU_DEVICE (int i, int j, int k) -> ReduceTuple
1156 {
1158 int ierr = coarsen_from_fine(AMREX_D_DECL(i,j,k), bx, 2,
1159 cvol,ccent,cba,cbc,cbn,
1160 AMREX_D_DECL(capx,capy,capz),
1161 AMREX_D_DECL(cfcx,cfcy,cfcz),
1162 AMREX_D_DECL(cecx[lidx],cecy[lidx],cecz[lidx]),
1163 cflag,fvol,fcent,fba,fbc,fbn,
1164 AMREX_D_DECL(fapx,fapy,fapz),
1165 AMREX_D_DECL(ffcx,ffcy,ffcz),
1166 AMREX_D_DECL(fecx[lidx],fecy[lidx],fecz[lidx]),
1167 fflag, face_dir);
1168 return {ierr};
1169 });
1170 }
1171
1172 ReduceTuple rv = reduce_data.value(reduce_op);
1173 error = amrex::max(0, amrex::get<0>(rv));
1174 }
1175
1176 {
1177 bool b = error;
1179 error = b;
1180 }
1181
1182 m_fc_data[face_dir]->m_built = true;
1183
1184 if (!error) {
1185 buildCellFlagFC(face_dir);
1186 }
1187}
1188
1189}
1190
1191#endif
Fixed-size array types for use on GPU and CPU.
#define BL_PROFILE(a)
Definition AMReX_BLProfiler.H:562
#define AMREX_ALWAYS_ASSERT_WITH_MESSAGE(EX, MSG)
Definition AMReX_BLassert.H:49
#define AMREX_ASSERT(EX)
Definition AMReX_BLassert.H:38
Problem-domain geometry: maps between index space and physical space.
#define AMREX_GPU_DEVICE
Definition AMReX_GpuQualifiers.H:18
amrex::ParmParse pp
Input file parser instance for the given namespace.
Definition AMReX_HypreIJIface.cpp:15
Array4< Real > fine
Definition AMReX_InterpFaceRegister.cpp:90
Array4< Real const > vfrac
Definition AMReX_MLEBNodeFDLaplacian.cpp:1468
GpuArray< Array4< Real const >, 3 > edgecent
Definition AMReX_MLEBNodeFDLaplacian.cpp:1557
#define AMREX_D_TERM(a, b, c)
Definition AMReX_SPACE.H:172
#define AMREX_D_DECL(a, b, c)
Definition AMReX_SPACE.H:171
void resize(const Box &b, int N=1, Arena *ar=nullptr)
This function resizes a BaseFab so it covers the Box b with N components.
Definition AMReX_BaseFab.H:1774
Array4< T const > array() const noexcept
Create an Array4 view over all components.
Definition AMReX_BaseFab.H:475
Elixir elixir() noexcept
Transfer owned storage to an Elixir when running in a GPU launch region.
Definition AMReX_BaseFab.H:1817
Reference-counted collection of Boxes.
Definition AMReX_BoxArray.H:681
bool coarsenable(int refinement_ratio, int min_width=1) const
Check whether every Box is coarsenable by refinement_ratio.
Definition AMReX_BoxArray.cpp:628
bool empty() const noexcept
Return whether the BoxArray is empty.
Definition AMReX_BoxArray.H:764
__host__ __device__ BoxND< new_dim > resize() const noexcept
Return a new BoxND of size new_dim by either shrinking or expanding this BoxND.
Definition AMReX_Box.H:927
__host__ __device__ bool contains(const IntVectND< dim > &p) const noexcept
Return true if argument is contained within BoxND.
Definition AMReX_Box.H:233
GpuArray< Real, 3 > CellSizeArray() const noexcept
Returns the cell sizes as a GpuArray for use on host or device.
Definition AMReX_CoordSys.H:85
Calculates the distribution of FABs to MPI processes.
Definition AMReX_DistributionMapping.H:51
static constexpr int ng
Definition AMReX_EB2_MultiGFab.H:26
Definition AMReX_EB2_Level.H:293
GShopLevel(IndexSpace const *is, const Geometry &geom)
Construct an empty level (usually regular) bound to geom.
Definition AMReX_EB2_Level.H:345
void define_fine(G const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Define data for the finest level using the supplied GeometryShop.
Definition AMReX_EB2_Level.H:480
void buildFCData(G const &gshop, int face_dir, int max_grid_size)
Build face-centered EB data for face_dir. Builds refined CC transiently.
Definition AMReX_EB2_Level.H:840
GShopLevel(IndexSpace const *is, int ilev, int max_grid_size, int ngrow, const Geometry &geom, GShopLevel< G > &fineLevel)
Construct a coarse level by coarsening fineLevel.
Definition AMReX_EB2_Level.H:799
GShopLevel(IndexSpace const *is, G const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Build a level directly from a GeometryShop object.
Definition AMReX_EB2_Level.H:350
static GShopLevel< G > makeAllRegular(IndexSpace const *is, const Geometry &geom)
Build a regular (no EB) level.
Definition AMReX_EB2_Level.H:325
void define_fine_mvmc(GS const &gshop, const Geometry &geom, int max_grid_size, int ngrow, bool extend_domain_face, int num_crse_opt)
Define a fine level for multi-valued multi-cut STL geometries (3-D only)
Definition AMReX_EB2_Level.H:764
Definition AMReX_EB2.H:84
Definition AMReX_EB2.H:28
Definition AMReX_EB2_Level.H:43
MultiFab m_bndryarea
Definition AMReX_EB2_Level.H:261
bool isOK() const noexcept
Definition AMReX_EB2_Level.H:49
void fillFaceCent(Array< MultiCutFab *, 3 > const &facecent, const Geometry &geom) const
Populate face centroids for each direction.
Definition AMReX_EB2_Level.cpp:794
MultiFab m_volfrac
Definition AMReX_EB2_Level.H:259
MultiGFab m_mgf
Definition AMReX_EB2_Level.H:256
bool m_allregular
Definition AMReX_EB2_Level.H:268
IntVect m_shift
Definition AMReX_EB2_Level.H:271
iMultiFab m_cutcellmask
Definition AMReX_EB2_Level.H:267
Array< MultiFab, 3 > m_facecent
Definition AMReX_EB2_Level.H:265
void fillBndryAreaFC(MultiFab &bndryarea, int face_dir, const Geometry &geom) const
Fill face-centered boundary areas for face_dir.
Definition AMReX_EB2_Level.cpp:1215
Array< MultiFab, 3 > m_edgecent
Definition AMReX_EB2_Level.H:266
void fillBndryNormFC(MultiFab &bndrynorm, int face_dir, const Geometry &geom) const
Fill face-centered boundary normals for face_dir.
Definition AMReX_EB2_Level.cpp:1251
void prepareForCoarsening(const Level &rhs, int max_grid_size, IntVect const &ngrow)
Prepare for coarsening by copying from rhs, respecting max grid size and grow vector.
Definition AMReX_EB2_Level.cpp:10
void fillCutCellMask(iMultiFab &cutcellmask, const Geometry &geom) const
Populate the cut-cell mask into cutcellmask.
Definition AMReX_EB2_Level.cpp:946
MultiFab m_centroid
Definition AMReX_EB2_Level.H:260
FabArray< EBCellFlagFab > m_cellflag
Definition AMReX_EB2_Level.H:258
const DistributionMapping & DistributionMap() const noexcept
Definition AMReX_EB2_Level.H:111
BoxArray m_covered_grids
Definition AMReX_EB2_Level.H:250
void fillLevelSet(MultiFab &levelset, const Geometry &geom) const
Write the implicit function values into levelset.
Definition AMReX_EB2_Level.cpp:907
Geometry m_geom
Definition AMReX_EB2_Level.H:247
Array< MultiFab, 3 > m_areafrac
Definition AMReX_EB2_Level.H:264
IntVect m_ngrow
Definition AMReX_EB2_Level.H:248
void fillEBCellFlagFC(FabArray< EBCellFlagFab > &cellflag, int face_dir, const Geometry &geom) const
Fill face-centered EBCellFlags for face_dir.
Definition AMReX_EB2_Level.cpp:1269
IndexSpace const * m_parent
Definition AMReX_EB2_Level.H:272
const FCData & getFCData(int face_dir) const
Get reference to face-centered data for face_dir (0=x, 1=y, 2=z).
Definition AMReX_EB2_Level.H:198
BoxArray m_grids
Definition AMReX_EB2_Level.H:249
void fillEdgeCentFC(Array< MultiFab *, 3 > const &edgecent, int face_dir, const Geometry &geom) const
Fill face-centered edge centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1353
const Geometry & Geom() const noexcept
Definition AMReX_EB2_Level.H:121
std::map< int, std::unique_ptr< MC::MCFab > > m_marching_cubes
Definition AMReX_EB2_Level.H:254
bool hasEBInfo() const noexcept
Definition AMReX_EB2_Level.H:138
void write_to_chkpt_file(const std::string &fname, bool extend_domain_face, int max_grid_size) const
Write this level’s EB data to a checkpoint file fname.
Definition AMReX_EB2_Level.cpp:956
void fillEBCellFlag(FabArray< EBCellFlagFab > &cellflag, const Geometry &geom) const
Fill cellflag with EBCellFlag data using geometry geom.
Definition AMReX_EB2_Level.cpp:429
void fillEdgeCent(Array< MultiCutFab *, 3 > const &edgecent, const Geometry &geom) const
Populate edge centroids for each direction.
Definition AMReX_EB2_Level.cpp:835
int coarsenFromFine(Level &fineLevel, bool fill_boundary)
Coarsen EB data from fineLevel into this level.
Definition AMReX_EB2_Level.cpp:84
MultiFab m_bndrynorm
Definition AMReX_EB2_Level.H:263
bool m_has_eb_info
Definition AMReX_EB2_Level.H:270
void fillAreaFrac(Array< MultiCutFab *, 3 > const &areafrac, const Geometry &geom) const
Populate face area fractions for each direction.
Definition AMReX_EB2_Level.cpp:640
MultiFab m_bndrycent
Definition AMReX_EB2_Level.H:262
MultiFab m_levelset
Definition AMReX_EB2_Level.H:257
void setShift(int direction, int ncells)
Shift this level by ncells cells along direction direction (for periodic tiling).
Definition AMReX_EB2_Level.cpp:1047
void buildCellFlagFC(int face_dir)
Build the face-centered EBCellFlag connectivity for face_dir from its area fractions.
Definition AMReX_EB2_Level.cpp:1062
bool isAllRegular() const noexcept
Definition AMReX_EB2_Level.H:47
static IntVect fcEdgeType(int idim, int face_dir) noexcept
Index type of the face-centered edge centroids for edge direction idim.
Definition AMReX_EB2_Level.H:185
void fillBndryCent(MultiCutFab &bndrycent, const Geometry &geom) const
Populate boundary centroids into bndrycent.
Definition AMReX_EB2_Level.cpp:592
void fillCentroidFC(MultiFab &centroid, int face_dir, const Geometry &geom) const
Fill face-centered centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1197
void fillBndryNorm(MultiCutFab &bndrynorm, const Geometry &geom) const
Populate boundary normals into bndrynorm.
Definition AMReX_EB2_Level.cpp:616
void fillBndryCentFC(MultiFab &bndrycent, int face_dir, const Geometry &geom) const
Fill face-centered boundary centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1233
MultiFab m_sdf
Definition AMReX_EB2_Level.H:252
IntVect const & nGrowVect() const noexcept
Definition AMReX_EB2_Level.H:126
void fillAreaFracFC(Array< MultiFab *, 3 > const &areafrac, int face_dir, const Geometry &geom) const
Fill face-centered area fractions for face_dir.
Definition AMReX_EB2_Level.cpp:1132
void fillVolFrac(MultiFab &vfrac, const Geometry &geom) const
Populate volume fractions into vfrac using geom.
Definition AMReX_EB2_Level.cpp:483
const BoxArray & boxArray() const noexcept
Definition AMReX_EB2_Level.H:109
Level(IndexSpace const *is, const Geometry &geom)
Construct a Level bound to EB index space is and Geometry geom.
Definition AMReX_EB2_Level.H:114
friend class GShopLevel
Definition AMReX_EB2_Level.H:278
bool m_ok
Definition AMReX_EB2_Level.H:269
void fillCentroid(MultiCutFab &centroid, const Geometry &geom) const
Populate cut-cell centroids into a MultiCutFab.
Definition AMReX_EB2_Level.cpp:544
Array< std::unique_ptr< FCData >, 3 > m_fc_data
Face-centered data storage: [0]=x-face, [1]=y-face, [2]=z-face.
Definition AMReX_EB2_Level.H:275
DistributionMapping m_dmap
Definition AMReX_EB2_Level.H:251
void fillBndryArea(MultiCutFab &bndryarea, const Geometry &geom) const
Populate boundary-face areas into bndryarea.
Definition AMReX_EB2_Level.cpp:568
void buildCellFlag()
Build EBCellFlag data from the current implicit function field.
Definition AMReX_EB2_Level.cpp:403
void buildCutCellMask(Level const &fine_level)
Build the cut-cell mask by coarsening from fine_level.
Definition AMReX_EB2_Level.cpp:966
bool hasFCData(int face_dir) const noexcept
Check if face-centered data is available for face_dir (0=x, 1=y, 2=z).
Definition AMReX_EB2_Level.H:192
IndexSpace const * getEBIndexSpace() const noexcept
Definition AMReX_EB2_Level.H:123
void fillVolFracFC(MultiFab &vfrac, int face_dir, const Geometry &geom) const
Fill face-centered volume fractions for face_dir.
Definition AMReX_EB2_Level.cpp:1093
void fillFaceCentFC(Array< MultiFab *, 3 > const &facecent, int face_dir, const Geometry &geom) const
Fill face-centered face centroids for face_dir.
Definition AMReX_EB2_Level.cpp:1321
LayoutData wrapper that allocates a GFab per FAB/Box.
Definition AMReX_EB2_MultiGFab.H:86
FAB storing EBCellFlag data per cell.
Definition AMReX_EBCellFlag.H:324
An Array of FortranArrayBox(FAB)-like Objects.
Definition AMReX_FabArray.H:356
Array4< typename FabArray< FAB >::value_type const > array(const MFIter &mfi) const noexcept
Read-only Array4 view for the FAB referenced by iterator mfi.
Definition AMReX_FabArray.H:621
Array4< typename FabArray< FAB >::value_type const > const_array(const MFIter &mfi) const noexcept
Synonym for array(const MFIter&) that highlights read-only semantics.
Definition AMReX_FabArray.H:649
Rectangular problem domain geometry.
Definition AMReX_Geometry.H:85
const Box & Domain() const noexcept
Returns our rectangular domain.
Definition AMReX_Geometry.H:244
Periodicity periodicity() const noexcept
Return the Periodicity based on the length of the domain.
Definition AMReX_Geometry.H:424
GpuArray< Real, 3 > ProbLoArray() const noexcept
Return the lo end of the problem domain in a GpuArray for device code.
Definition AMReX_Geometry.H:217
bool isPeriodic(int dir) const noexcept
Is the domain periodic in the specified direction?
Definition AMReX_Geometry.H:397
Definition AMReX_GpuElixir.H:13
__host__ static __device__ constexpr IntVectND< dim > TheCellVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:810
__host__ static __device__ constexpr IntVectND< dim > TheZeroVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:771
__host__ static __device__ constexpr IntVectND< dim > TheNodeVector() noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:801
__host__ static __device__ constexpr IntVectND< dim > TheDimensionVector(int d) noexcept
This static member function returns a reference to a constant IntVectND object, all of whose dim argu...
Definition AMReX_IntVect.H:790
Iterator for looping ever tiles and boxes of amrex::FabArray based containers.
Definition AMReX_MFIter.H:88
bool isValid() const noexcept
Is the iterator valid i.e. is it associated with a FAB?
Definition AMReX_MFIter.H:176
FabArray façade that only allocates cut-cell Fabs (skipping regular cells).
Definition AMReX_MultiCutFab.H:87
A collection (stored as an array) of FArrayBox objects.
Definition AMReX_MultiFab.H:40
void define(const BoxArray &bxs, const DistributionMapping &dm, int nvar, int ngrow, const MFInfo &info=MFInfo(), const FabFactory< FArrayBox > &factory=FArrayBoxFactory())
Definition AMReX_MultiFab.cpp:519
Parse Parameters From Command Line and Input Files.
Definition AMReX_ParmParse.H:353
int queryAdd(std::string_view name, T &ref)
If name is found, the value in the ParmParse database will be stored in the ref argument....
Definition AMReX_ParmParse.H:1049
This class provides the user with a few print options.
Definition AMReX_Print.H:35
Definition AMReX_Reduce.H:438
Type value()
Definition AMReX_Reduce.H:473
Definition AMReX_Reduce.H:597
void eval(MF const &mf, IntVect const &nghost, D &reduce_data, F &&f)
Definition AMReX_Reduce.H:734
A Collection of IArrayBoxes.
Definition AMReX_iMultiFab.H:34
amrex_real Real
Floating Point Type for Fields.
Definition AMReX_REAL.H:80
amrex_long Long
Definition AMReX_INT.H:30
__host__ __device__ BoxND< dim > convert(const BoxND< dim > &b, const IntVectND< dim > &typ) noexcept
Return a copy of b converted to the nodal flags typ.
Definition AMReX_Box.H:1630
__host__ __device__ BoxND< dim > surroundingNodes(const BoxND< dim > &b, int dir) noexcept
Return a BoxND with NODE based coordinates in direction dir that encloses BoxND b.
Definition AMReX_Box.H:1582
__host__ __device__ BoxND< dim > coarsen(const BoxND< dim > &b, int ref_ratio) noexcept
Return a copy of b coarsened by the isotropic ratio ref_ratio.
Definition AMReX_Box.H:1469
__host__ __device__ BoxND< dim > enclosedCells(const BoxND< dim > &b, int dir) noexcept
Return a BoxND with CELL based coordinates in direction dir that is enclosed by b.
Definition AMReX_Box.H:1664
__host__ __device__ BoxND< dim > grow(const BoxND< dim > &b, int i) noexcept
Return a copy of b grown uniformly by i cells in every direction.
Definition AMReX_Box.H:1326
__host__ __device__ BoxND< dim > refine(const BoxND< dim > &b, int ref_ratio) noexcept
Return a copy of b refined by the isotropic ratio ref_ratio.
Definition AMReX_Box.H:1510
std::array< T, N > Array
Definition AMReX_Array.H:31
void Or(bool &v, MPI_Comm comm)
Definition AMReX_ParallelReduce.H:236
__host__ __device__ constexpr const T & max(const T &a, const T &b) noexcept
Definition AMReX_Algorithm.H:53
Definition AMReX_FabArrayBase.H:38
void streamSynchronize() noexcept
Definition AMReX_GpuDevice.H:310
bool inLaunchRegion() noexcept
Definition AMReX_GpuControl.H:88
bool notInLaunchRegion() noexcept
Definition AMReX_GpuControl.H:89
void write_stl(std::string const &filename, std::map< int, std::unique_ptr< MCFab > > const &mc_fabs)
Write the collected marching-cubes output to an STL file.
Definition AMReX_MarchingCubes.cpp:925
void marching_cubes(Geometry const &geom, FArrayBox &sdf_fab, MCFab &mc_fab)
Run marching cubes on signed-distance field sdf_fab.
Definition AMReX_MarchingCubes.cpp:693
void Initialize()
Initialize internal lookup tables and device buffers for marching cubes.
Definition AMReX_MarchingCubes.cpp:651
MPI_Comm CommunicatorSub() noexcept
sub-communicator for current frame
Definition AMReX_ParallelContext.H:70
int MyProcSub() noexcept
my sub-rank in current frame
Definition AMReX_ParallelContext.H:76
int NProcsSub() noexcept
number of ranks in current frame
Definition AMReX_ParallelContext.H:74
Definition AMReX_Amr.cpp:50
__host__ __device__ void ignore_unused(const Ts &...)
No-op helper that marks variables as intentionally unused.
Definition AMReX.H:259
RunOn
Definition AMReX_GpuControl.H:65
void AllGatherBoxes(Vector< Box > &bxs, int n_extra_reserve)
Gather boxes from all MPI ranks into bxs.
Definition AMReX_Box.cpp:124
BoxND< 3 > Box
Box is an alias for amrex::BoxND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:35
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
__host__ __device__ constexpr IntVectND< dim > scale(const IntVectND< dim > &p, int s) noexcept
Returns a IntVectND obtained by multiplying each of the components of this IntVectND by s.
Definition AMReX_IntVect.H:1106
int Verbose() noexcept
Return the verbosity level configured via ParmParse or SetVerbose().
Definition AMReX.cpp:182
void Abort(const std::string &msg)
Print a fatal-error message to stderr and abort execution.
Definition AMReX.cpp:242
void LoopOnCpu(Dim3 lo, Dim3 hi, F const &f) noexcept
Definition AMReX_Loop.H:365
A multidimensional array accessor.
Definition AMReX_Array4.H:289
Face-centered EB data storage structure.
Definition AMReX_EB2_Level.H:150
bool m_built
Definition AMReX_EB2_Level.H:174
Array< MultiFab, 3 > m_areafrac_fc
Definition AMReX_EB2_Level.H:157
MultiFab m_bndryarea_fc
Definition AMReX_EB2_Level.H:154
Array< MultiFab, 3 > m_edgecent_fc
Definition AMReX_EB2_Level.H:159
BoxArray m_covered_grids_fc
Definition AMReX_EB2_Level.H:173
MultiFab m_bndrycent_fc
Definition AMReX_EB2_Level.H:155
MultiFab m_centroid_fc
Definition AMReX_EB2_Level.H:153
MultiFab m_bndrynorm_fc
Definition AMReX_EB2_Level.H:156
MultiFab m_volfrac_fc
Definition AMReX_EB2_Level.H:152
Array< MultiFab, 3 > m_facecent_fc
Definition AMReX_EB2_Level.H:158
FabArray< EBCellFlagFab > m_cellflag_fc
Definition AMReX_EB2_Level.H:151
FabArray memory allocation information.
Definition AMReX_FabArray.H:73
MFInfo & SetTag() noexcept
Terminate a variadic SetTag() call.
Definition AMReX_FabArray.H:90
MFInfo & SetAlloc(bool a) noexcept
Control whether FAB storage is allocated when the FabArray is defined.
Definition AMReX_FabArray.H:81
Definition AMReX_MFIter.H:20
MFItInfo & SetDynamic(bool f) noexcept
Definition AMReX_MFIter.H:43